/************************************************************************/
/* TinyTalk keyboard handler.						 */
/* */
/* Version 1.0 [ 1/24/90] : Initial implementation by ABR.		 */
/* 1.1 [ 1/25/90] : Added minimal termcap support.		 */
/* 1.2 [ 1/26/90] : Clean up terminal state on ^C.		 */
/* */
/************************************************************************/

#include "TinyMUD_copyright.h"
#include <signal.h>
#include <sgtty.h>
#include <errno.h>
#include <stdio.h>

#include "client.h"

/* For some odd systems, which don't put this in errno.h. */

extern int      errno;

#ifdef TERMCAP
extern char    *getenv();
static int      have_clear;
static char     start_of_line[16], clear_to_eol[16];	/* Warning, short! */
static char     write_buffer[64];	/* Also rather short.... */
static char    *write_buffer_ptr;
#endif
int		have_inv = 0;

static int      num_columns;
static int      current_column;
static string   keyboard_buffer;
static int      keyboard_pos;

sigint_handler()
{				/* Restore terminal state; cleanup. */
	die("\n\nInterrupt, exiting.\n");    /* Exit through error handler. */
}

sigtstp_handler()
{				/* Leave terminal alone! */
	cooked_echo_mode();
	kill(getpid(), SIGSTOP);
	cbreak_noecho_mode();

	if (keyboard_pos != 0)		/* If there is input, we want to  */
		set_refresh_pending();	/* refresh when we start back up. */
}

init_keyboard()
{
	struct winsize  size;

	cbreak_noecho_mode();

	ioctl(0, TIOCGWINSZ, &size);
	num_columns = size.ws_col;
	if (num_columns <= 20)	/* Suspicious, reset it. */
		num_columns = 80;
	set_default_wrap(num_columns);

	keyboard_pos = 0;	/* No input yet. */
	current_column = 0;	/* Left side of screen. */

	signal(SIGINT, sigint_handler);		/* Control-C. */
	signal(SIGTSTP, sigtstp_handler);	/* Control-Z. */

#ifdef TERMCAP
	get_termcap_info();
#endif
}

#ifdef TERMCAP

get_termcap_info()
{
	char            blob[1024];	/* As per termcap man page */
	char           *terminal_name, *temp;

	terminal_name = getenv("TERM");
	if (terminal_name == NULL) {
		have_clear = FALSE;
		return;
	}
	if (tgetent(blob, terminal_name) == 1) {
		temp = start_of_line;
		if (tgetstr("cr", &temp) == NULL) {
			start_of_line[0] = '\r';
			start_of_line[1] = '\0';
		}
		temp = clear_to_eol;
		if (tgetstr("ce", &temp) == NULL)
			have_clear = FALSE;
		else
			have_clear = TRUE;
		if (tgetstr("so", &temp) == NULL)
			have_inv = 0;
		else
			have_inv = 1;
	} else			/* Couldn't read termcap entry */
		have_clear = FALSE;
}

#endif

cleanup_keyboard()
{
	cooked_echo_mode();
	signal(SIGTSTP, NULL);
}

handle_keyboard_input()
{				/* Read input, one char at a time. */
	char            ch;
	int             count;

	count = read(0, &ch, 1);/* Read from stdin. */
	if (count == 0)		/* Huh?  Ignore this. */
		return;

	if (count == -1) {
		if (errno == EWOULDBLOCK)
			return;
		else
			die("%% Couldn't read keyboard.\n");
	}
	if (ch == '\n') {
		write(1, &ch, 1);
		process_buffer();
		keyboard_pos = 0;
		current_column = 0;
	} else {
		if ((ch == '\010') || (ch == '\177')) {	/* Backspace. */
			if (keyboard_pos == 0)
				return;

			keyboard_pos--;

			if (current_column != 0) {    /* Can do BS, SPC, BS. */
				write(1, "\010 \010", 3);
				current_column--;
			} else {
				current_column = num_columns - 1;
						/* Wrapped around. */
				do_refresh();	
					/* Will redraw the right part. */
			}
		} else if ((ch == '\025') || (ch == '\030')) {	/* Cancel line. */
			keyboard_pos = 0;
			erase_keyboard_input();
			current_column = 0;
		} else if ((ch < ' ') || (keyboard_pos == MAXSTRLEN - 1)) {
			ch = '\007';	/* Beep! */
			write(1, &ch, 1);
		} else {
			keyboard_buffer[keyboard_pos++] = ch;
			current_column = (current_column + 1) % num_columns;
			write(1, &ch, 1);
		}
	}
}

process_buffer()
{
	if (keyboard_pos == 0)	/* Nothing typed. */
		return;

	keyboard_buffer[keyboard_pos] = '\n';
	transmit(keyboard_buffer, keyboard_pos + 1);
/*	log_input(keyboard_buffer);*/
}

do_refresh()
{
	int             loc;	/* Place to start refresh from */
	int             modulo;

	modulo = keyboard_pos % num_columns;	/* Number on last line */
	loc = keyboard_pos - modulo;

	if (modulo != 0)
		write(1, &keyboard_buffer[loc], modulo);

	clear_refresh_pending();
}

#ifdef TERMCAP
write_one(c)
	char            c;
{
	*(write_buffer_ptr++) = c;
}
#endif

erase_keyboard_input()
{
#ifdef TERMCAP

	if (have_clear) {	/* Can we use termcap now? */
		write_buffer_ptr = write_buffer;
		tputs(start_of_line, 1, write_one);
		tputs(clear_to_eol, 1, write_one);
		write(1, write_buffer, write_buffer_ptr - write_buffer);
		goto end_termcap;
	}
	/* Drops through into non-TERMCAP code otherwise... */

#endif

	{
		int             temp;
		string          buffer;	/* Should be long enough for 80*3. */
		char           *ptr;

		temp = current_column;
		ptr = buffer;

		while (temp != 0) {	/* Erase line, backwards. */
			*(ptr++) = '\010';
			*(ptr++) = ' ';
			*(ptr++) = '\010';
			temp--;
		}
		write(1, buffer, ptr - buffer);
	}
#ifdef TERMCAP
end_termcap:
#endif TERMCAP

	if (keyboard_pos != 0)	/* If there is input, schedule a */
		set_refresh_pending();	/* refresh event. */
}
